import os

MAXTIME = 3600

TABLEHEADER_FULL = r"""\begin{table}
  \caption{Test results setting \emph{%s}.}
    \label{tab:results%s}
    \begin{scriptsize}
    \begin{tabular*}{\columnwidth}{@{}l@{\;\;\extracolsep{\fill}}rrrrrrrr@{}}\toprule"""

TABLEHEADER_SUMM = r"""\begin{table}
  \caption{Test results settings.}
    \label{tab:results_summary}
    \begin{scriptsize}
    \begin{tabular*}{\columnwidth}{@{}l@{\;\;\extracolsep{\fill}}rrrr@{}}\toprule"""

TABLEFOOTER = r"""\bottomrule
    \end{tabular*}
    \end{scriptsize}
\end{table}
"""


def calcShiftedGeomean(x, shift):
    shiftedtimegeom = 1
    for i, xx in enumerate(x):
        shiftedtimegeom = (shiftedtimegeom ** (float(i)/(i+1))) * (max(xx+shift, 1.0)**(1.0/(i+1)))
    return shiftedtimegeom - shift


def readOutGetTimeConsKResilient(f):
    with open(f, 'r') as f:
        lines = f.readlines()
        seen = False
        for line in lines:
            if "Constraint Timings" in line:
                seen = True
            if "kresilient       :" in line and seen:
                line = line.split()
                return float(line[2])
    return 0


def readOutGetNrCalledConsKResilient(f):
    with open(f, 'r') as f:
        lines = f.readlines()
        seen = False
        for line in lines:
            if "Constraints        :    " in line:
                seen = True
            if "kresilient       :" in line and seen:
                line = line.split()
                return int(line[4]) + int(line[6])
    return 0


def readOutGetNrCutsConsKResilient(f):
    with open(f, 'r') as f:
        lines = f.readlines()
        seen = False
        for line in lines:
            if "Constraints        :    " in line:
                seen = True
            if "kresilient       :" in line and seen:
                line = line.split()
                return int(line[13]) + int(line[15])
    return 0


def readOutTotalSolvingTime(f):
    with open(f, 'r') as f:
        lines = f.readlines()
        for line in lines:
            if "Total Time         :" in line:
                line = line.split(":")
                return float(line[1])
            elif "not sufficient time or memory to construct the whole scenario formulation" in line:
                return float("inf")
    return 0


def readOutInterdictionSolvingTime(f):
    with open(f, 'r') as f:
        lines = f.readlines()
        for line in reversed(lines):
            if "Total time scipint solving" in line:
                line = line.split("solving")
                return float(line[1])
    return 0


def readOutScipfeasSolvingTime(f):
    with open(f, 'r') as f:
        lines = f.readlines()
        for line in reversed(lines):
            if "Total time subscip solving" in line:
                line = line.split("solving")
                return float(line[1])
    return 0


def readOutInterdictionNodes(f):
    with open(f, 'r') as f:
        lines = f.readlines()
        for line in reversed(lines):
            if "Total nodes scipint solving " in line:
                line = line.split("solving")
                return int(line[1])
    return 0


def readOutScipfeasNodes(f):
    with open(f, 'r') as f:
        lines = f.readlines()
        for line in reversed(lines):
            if "Total nodes subscip solving " in line:
                line = line.split("solving")
                return int(line[1])
    return 0


def readOutTotalNodes(f):
    n = readOutScipfeasNodes(f) + readOutInterdictionNodes(f)
    with open(f, 'r') as f:
        lines = f.readlines()
        for line in reversed(lines):
            if "nodes            :" in line:
                line = line.split(":")[1].split("(")
                return int(line[0]) + n
    return n


def readOutInterdictionNSolves(f):
    with open(f, 'r') as f:
        lines = f.readlines()
        for line in reversed(lines):
            if "scipint solves " in line:
                line = line.split("solves")
                return int(line[1])
    return 0


def readOutScipfeasNSolves(f):
    with open(f, 'r') as f:
        lines = f.readlines()
        for line in reversed(lines):
            if "subscip solves " in line:
                line = line.split("solves")
                return int(line[1])
    return 0


def readOutScipfeasSolutions(f):
    with open(f, 'r') as f:
        lines = f.readlines()
        for line in reversed(lines):
            if "separated scip_feas solutions " in line:
                line = line.split("solutions")
                return int(line[1])
    return 0


def readOutDualBound(f):
    with open(f, 'r') as f:
        lines = f.readlines()
        for line in reversed(lines):
            if "Dual Bound       :" in line:
                if "Dual Bound       :          -" in line:
                    return -float("inf")
                line = line.split(":")
                return float(line[1])
            elif "not sufficient time or memory to construct the whole scenario formulation" in line:
                return -float("inf")

    raise Exception("did not find dual bound line, file %s" % f)


def readOutPrimalBound(f):
    with open(f, 'r') as f:
        lines = f.readlines()
        for line in reversed(lines):
            if "  Primal Bound     :" in line:
                if "Primal Bound     :          -" in line:
                    return float("inf")
                if "Primal Bound     : infeasible" in line:
                    return -float("inf")
                line = line.split(":")[1]
                return float(line.split("(")[0])
            elif "not sufficient time or memory to construct the whole scenario formulation" in line:
                return float("inf")
    raise Exception("did not find primal bound line")


def readOutSolved(f):
    primal = readOutPrimalBound(f)
    dual = readOutDualBound(f)
    solved = primal in (dual, -float("inf"))
    time = readOutTotalSolvingTime(f)

    assert solved != (time >= MAXTIME)
    return solved


def readOutScipfeasTimePerSolve(f):
    if readOutScipfeasNSolves(f) > 0:
        return readOutScipfeasSolvingTime(f) / readOutScipfeasNSolves(f)
    return 0.0


def printBoundsInstances(tset, settings):
    """Given a tset and a list of settings go through all .out files in
       RESULT_DIR and print for each setting a table containing primal and dual bound.
    """
    testsetlen = -1
    for setting in settings.split():
        print(TABLEHEADER_FULL % (setting.replace("_", "\\_"), setting))
        print(r"instance & time & primal & dual &\\ ")

        timesInterdiction = []
        timesScipfeas = []
        nrConsKCutAdded = []
        nrConsInterdictionCutAdded = []
        nrConsKCalled = []
        timesTotal = []
        nrNodesTotal = []
        nrNodesInterdiction = []
        nrNodesScipfeas = []
        dualbound = []
        primalbound = []
        nrSolvesInterdiction = []
        nrSolvesScipfeas = []
        solved = []
        primal = []
        dual = []
        timePerSolveScipfeas = []

        outfiles = sorted(list(filter(lambda x: ".out" in x and tset in x and setting in x and "check" not in x, FILES)), key=lambda x: int(x.split(".")[2].split("_")[0]))
        if testsetlen == -1:
            testsetlen = len(outfiles)
        for f in outfiles:
            f = RESULT_DIR + f
            timesInterdiction += [min(readOutInterdictionSolvingTime(f), MAXTIME)]
            timesScipfeas += [min(readOutScipfeasSolvingTime(f), MAXTIME)]
            nrConsKCutAdded += [readOutGetNrCutsConsKResilient(f)]
            nrConsInterdictionCutAdded += [readOutScipfeasSolutions(f)]
            timesTotal += [min(readOutTotalSolvingTime(f), MAXTIME)]
            nrNodesTotal += [readOutTotalNodes(f)]
            nrNodesInterdiction += [readOutInterdictionNodes(f)]
            nrNodesScipfeas += [readOutScipfeasNodes(f)]
            dualbound = [readOutDualBound(f)]
            primalbound = [readOutPrimalBound(f)]
            nrSolvesInterdiction += [readOutInterdictionNSolves(f)]
            nrSolvesScipfeas += [readOutScipfeasNSolves(f)]
            solved += [readOutSolved(f)]
            primal += [readOutPrimalBound(f)]
            dual += [readOutDualBound(f)]
            timePerSolveScipfeas += [readOutScipfeasTimePerSolve(f)]

            print("\\verb {0} & {1:.2f} & {2:.5f} & {3:.5f} \\\\".format(f.split(".")[2], timesTotal[-1], primal[-1], dual[-1]))
        print(TABLEFOOTER)
        print("\n\n")


def printTablesAllInstances(tset, settings):
    """Given a tset and a list of settings go through all .out files in
       RESULT_DIR and print for each setting a table containing relevant instance data.
    """
    testsetlen = -1
    for setting in settings.split():
        print(TABLEHEADER_FULL % (setting.replace("_", "\\_"), setting))
        print(r"instance & time & time $\Failset$ & time $\SetMinFunc$ & \# nodes & \# nodes $\Failset$ & \# nodes $\SetMinFunc$ & \# solves $\Failset$ & \# solves $\SetMinFunc$ & $\card{\Failset}$ & $\card{\SetMinFunc}$ & solved\\")
        print(r"\midrule")

        timesInterdiction = []
        timesScipfeas = []
        nrConsKCutAdded = []
        nrConsInterdictionCutAdded = []
        nrConsKCalled = []
        timesTotal = []
        nrNodesTotal = []
        nrNodesInterdiction = []
        nrNodesScipfeas = []
        dualbound = []
        primalbound = []
        nrSolvesInterdiction = []
        nrSolvesScipfeas = []
        solved = []

        outfiles = sorted(list(filter(lambda x: ".out" in x and tset in x and setting in x and "check" not in x, FILES)), key=lambda x: int(x.split(".")[2].split("_")[0]))
        if testsetlen == -1:
            testsetlen = len(outfiles)
        for f in outfiles:
            f = RESULT_DIR + f
            timesInterdiction += [min(readOutInterdictionSolvingTime(f), MAXTIME)]
            timesScipfeas += [min(readOutScipfeasSolvingTime(f), MAXTIME)]
            nrConsKCutAdded += [readOutGetNrCutsConsKResilient(f)]
            nrConsInterdictionCutAdded += [readOutScipfeasSolutions(f)]
            timesTotal += [min(readOutTotalSolvingTime(f), MAXTIME)]
            nrNodesTotal += [readOutTotalNodes(f)]
            nrNodesInterdiction += [readOutInterdictionNodes(f)]
            nrNodesScipfeas += [readOutScipfeasNodes(f)]
            dualbound = [readOutDualBound(f)]
            primalbound = [readOutPrimalBound(f)]
            nrSolvesInterdiction += [readOutInterdictionNSolves(f)]
            nrSolvesScipfeas += [readOutScipfeasNSolves(f)]
            solved += [readOutSolved(f)]

            print("\\verb {0} & {1:.2f} & {2:.2f} & {3:.2f} & {4} & {5} & {6} & {7} & {8} & {9} & {10} & {11} \\\\".format(f.split(".")[2], timesTotal[-1], timesInterdiction[-1], timesScipfeas[-1], nrNodesTotal[-1], nrNodesInterdiction[-1], nrNodesScipfeas[-1], nrSolvesInterdiction[-1], nrSolvesScipfeas[-1], nrConsKCutAdded[-1], nrConsInterdictionCutAdded[-1], "\\checkmark" if solved[-1] else "--"))
        print(r"\midrule")
        print("sh. geo. mean [1,100]    & {0:.2f}     &         &         & {1:.2f} &  & & & & & & {2}\\\\".format(calcShiftedGeomean(timesTotal, 1), calcShiftedGeomean(nrNodesTotal, 100), sum(solved)))
        print(TABLEFOOTER)
        print("\n\n")


def printTablesSummary(tset, settings):
    """Given a testset tset and a list of settings go through all .out files in
       RESULT_DIR and print a table summarizing for each setting the overall results
    """
    testsetlen = -1
    print(TABLEHEADER_SUMM)
    print(r"""\multicolumn{2}{c}{cut improvement strategy}     & \multirow{2}{*}{time} & \multirow{2}{*}{\# nodes} &  \multirow{2}{*}{\# solved} \\
            \cmidrule{1-2} greedy & resolving & \\
            \midrule""")

    for setting in settings.split():

        timesInterdiction = []
        timesScipfeas = []
        nrConsKCutAdded = []
        nrConsInterdictionCutAdded = []
        nrConsKCalled = []
        timesTotal = []
        nrNodesTotal = []
        nrNodesInterdiction = []
        nrNodesScipfeas = []
        dualbound = []
        primalbound = []
        nrSolvesInterdiction = []
        nrSolvesScipfeas = []
        solved = []

        outfiles = sorted(list(filter(lambda x: ".out" in x and tset in x and setting in x and "check" not in x, FILES)), key=lambda x: int(x.split(".")[2].split("_")[0]))
        if testsetlen == -1:
            testsetlen = len(outfiles)
        for f in outfiles:
            f = RESULT_DIR + f
            timesInterdiction += [min(readOutInterdictionSolvingTime(f), MAXTIME)]
            timesScipfeas += [min(readOutScipfeasSolvingTime(f), MAXTIME)]
            nrConsKCutAdded += [readOutGetNrCutsConsKResilient(f)]
            nrConsInterdictionCutAdded += [readOutScipfeasSolutions(f)]
            timesTotal += [min(readOutTotalSolvingTime(f), MAXTIME)]
            nrNodesTotal += [readOutTotalNodes(f)]
            nrNodesInterdiction += [readOutInterdictionNodes(f)]
            nrNodesScipfeas += [readOutScipfeasNodes(f)]
            dualbound = [readOutDualBound(f)]
            primalbound = [readOutPrimalBound(f)]
            nrSolvesInterdiction += [readOutInterdictionNSolves(f)]
            nrSolvesScipfeas += [readOutScipfeasNSolves(f)]
            solved += [readOutSolved(f)]
        setUsedGreedy = "--"
        if "UseIncCov" in setting:
            setUsedGreedy = r"$\nearrow$"
        if "UseDecCov" in setting:
            setUsedGreedy = r"$\searrow$"
        setUsedResolve = "--"
        if "ResolveNCuts_inf" in setting:
            setUsedResolve = r"$\infty$"
        elif "ResolveNCuts" in setting:
            setUsedResolve = setting.split("ResolveNCuts_")[1]
            setUsedResolve = setUsedResolve.split("_")[0]

        usedSetting = " {0} & {1}".format(setUsedGreedy, setUsedResolve)
        if "NoConsRes" in setting:
            usedSetting = r"\multicolumn{2}{c}{extended formulation}"

        print(r" {0}    & {1:.2f}     & {2:.2f} & {3} \\".format(usedSetting, calcShiftedGeomean(timesTotal, 1), calcShiftedGeomean(nrNodesTotal, 100), sum(solved)))
    print(TABLEFOOTER)


def compareTestWithSolu(tset, settings):
    """Given a testset tset and a list of settings go through all .out files in
       RESULT_DIR and compare the result to the .solu files which are assumed to
       lie in testset/ .
    """
    instancessolu = {}
    # read solu file
    with open("testset/" + tset + ".solu", 'r') as f:
        lines = f.readlines()
        for line in reversed(lines):
            line = line.split()
            if "inf" not in line[0]:
                instancessolu[line[1]] = (line[0], float(line[2]))
            else:
                instancessolu[line[1]] = (line[0])
    # compare for each setting and file

    for setting in settings.split():
        outfiles = sorted(list(filter(lambda x: ".out" in x and tset in x and setting in x and "check" not in x, FILES)), key=lambda x: int(x.split(".")[2].split("_")[0]))
        for f in outfiles:
            instance = f.split(".")[2]
            instance = instance[instance.find("_")+1:]
            if instance not in instancessolu:
                print("Can't find %s in solu file." % instance)
            solu = instancessolu[instance]
            f = result_dir + f
            # solved = readOutSolved(f)
            primal = readOutPrimalBound(f)
            dual = readOutDualBound(f)
            if solu[0] == "=opt=":
                if solu[1] - primal > 0.0001:
                    raise Exception("instance %s, setting %s has a better value then solu." % (instance, setting))
                if dual - solu[1] > 0.0001:
                    raise Exception("instance %s, setting %s cuts off solu." % (instance, setting))
            elif solu[0] == "=best=":
                if primal < solu[1]:
                    print("instance %s, setting %s has a better value then solu." % (instance, setting))
                if dual > solu[1]:
                    raise Exception("instance %s, setting %s cuts off solu." % (instance, setting))
            else:
                if -float("inf") < primal < float("inf"):
                    raise Exception("instance %s, setting %s finds a solu." % (instance, setting))


#
#
# Program start
#
#

SETTINGS = """inexact_ConsRes_NoExtended_NoRand_NoResolve_NoReopt_FixBds
inexact_ConsRes_NoExtended_NoRand_ResolveNCuts_5_NoReopt_FixBds
inexact_ConsRes_NoExtended_NoRand_ResolveNCuts_inf_NoReopt_FixBds
inexact_ConsRes_NoExtended_NoRand_UseDecCov_NoResolve_NoReopt_FixBds
inexact_ConsRes_NoExtended_NoRand_UseDecCov_ResolveNCuts_5_NoReopt_FixBds
inexact_ConsRes_NoExtended_NoRand_UseDecCov_ResolveNCuts_inf_NoReopt_FixBds
inexact_ConsRes_NoExtended_NoRand_UseIncCov_NoResolve_NoReopt_FixBds
inexact_ConsRes_NoExtended_NoRand_UseIncCov_ResolveNCuts_5_NoReopt_FixBds
inexact_ConsRes_NoExtended_NoRand_UseIncCov_ResolveNCuts_inf_NoReopt_FixBds
inexact_ConsRes_Extended_NoRand_NoResolve_NoReopt_FixBds
inexact_ConsRes_Extended_NoRand_ResolveNCuts_5_NoReopt_FixBds
inexact_ConsRes_Extended_NoRand_ResolveNCuts_inf_NoReopt_FixBds
inexact_NoConsRes"""

TESTSETS = "truss_buckling_noLap"
TESTSETS = "hr_testrig"
TESTSETS = "gas"

# folder in which the outfiles belong to
RESULT_DIR = "outfiles/"
FILES = os.listdir(RESULT_DIR)

for testset in TESTSETS.split():
    print(testset)
#    compareTestWithSolu(testset, SETTINGS)
    printTablesAllInstances(testset, SETTINGS)
#    printBoundsInstances(testset, SETTINGS)
    printTablesSummary(testset, SETTINGS)
